import java.text.SimpleDateFormat
import com.google.common.collect.ArrayListMultimap


ganttChart = new GanttChartTool(issues, startDateCustomFieldId, endDateCustomFieldId, progressPercentageCustomFieldId, viewPeriod, componentAccessor)

/**
 * Velocity tool to help rendering Gantt charts with Better PDF Exporter for Jira Cloud.
 * @see gantt-chart-fo.vm as the actual template.
 */
public class GanttChartTool {
	def componentAccessor

	def startDateCustomFieldFullId
	def endDateCustomFieldFullId
	def progressPercentageCustomFieldFullId

	/**
	 * The dates to iterate over (defined by the time period spanned by the tasks).
	 */
	def dates = new ArrayList()
	def earliestDate = null
	def latestDate = null

	def earliestDateDayOfWeek = -1
	def currentDateIndex = -1
	long millisPerUnitPeriod
	def actualViewPeriod

	def hasSubTasks
	def subTasksByParentId = ArrayListMultimap.create()

	static final calendar = Calendar.instance
	static final monthOnlyDateFormat = new SimpleDateFormat("MMM")
	static final dayOnlyDateFormat = new SimpleDateFormat("d")

	static final VIEW_PERIOD_DAY = "day"
	static final VIEW_PERIOD_WEEK = "week"
	static final VIEW_PERIOD_AUTO = "auto"

	/**
	 * Periods longer than this (in unit of days) will default to "week" resolution.
	 */
	static final int AUTO_VIEW_PERIOD_DAY_LIMIT = 31
	static final long MILLIS_PER_DAY = 24*60*60*1000
	static final long MILLIS_PER_WEEK = 7*MILLIS_PER_DAY

	public GanttChartTool(issues, startDateCustomFieldId, endDateCustomFieldId, progressPercentageCustomFieldId, viewPeriod, componentAccessor) {
		this.componentAccessor = componentAccessor

		startDateCustomFieldFullId = getCustomFieldName(startDateCustomFieldId)
		endDateCustomFieldFullId = getCustomFieldName(endDateCustomFieldId)
		progressPercentageCustomFieldFullId = getCustomFieldName(progressPercentageCustomFieldId)

		calendar.clear()
		calendar.firstDayOfWeek = Calendar.MONDAY

		// find the earliest and latest dates for fully defined tasks
		for(issue in issues) {
			def startDate = getIssueStartDate(issue)
			def endDate = getIssueEndDate(issue)
			if(startDate != null) {
				if((earliestDate == null) || startDate.before(earliestDate)) {
					earliestDate = startDate
				}
			}

			if(endDate != null) {
				if((latestDate  == null) || endDate.after(latestDate)) {
					latestDate = endDate
				}
			}

			// build issue to subtasks mapping
			if(issue.isSubTask()) {
				hasSubTasks = true
				subTasksByParentId.put(issue.parentId, issue)
			}
		}

		// select the optimal period automatically
		if(viewPeriod == VIEW_PERIOD_AUTO) {
			millisPerUnitPeriod = MILLIS_PER_DAY
			actualViewPeriod = (getLength() <= AUTO_VIEW_PERIOD_DAY_LIMIT) ? VIEW_PERIOD_DAY : VIEW_PERIOD_WEEK
		} else {
			actualViewPeriod = viewPeriod
		}

		// generate the dates spanned
		millisPerUnitPeriod = (actualViewPeriod == VIEW_PERIOD_DAY) ? MILLIS_PER_DAY : MILLIS_PER_WEEK
		if((earliestDate != null) && (latestDate != null)) {
			earliestDate = trimToDay(earliestDate, Calendar.MONDAY)
			latestDate = trimToDay(latestDate, Calendar.SUNDAY)

			Date date = earliestDate
			def i = 0
			while(!date.after(latestDate)) {
				dates.add(date)
				i += 1
				date = new Date(earliestDate.time + i*millisPerUnitPeriod)
			}

			calendar.setTime(earliestDate)
			earliestDateDayOfWeek = calendar.get(Calendar.DAY_OF_WEEK)

			def currentDate = trimToMidnight(new Date())
			currentDateIndex = (int)((currentDate.time - earliestDate.time) / millisPerUnitPeriod)
		}
	}

	private static String getCustomFieldName(customFieldId) {
		return ((customFieldId instanceof Integer) && !customFieldId.equals(0)) ? "customfield_${customFieldId}" : null
	}

	public String getStartDateLabel() {
		return (startDateCustomFieldFullId != null) ?
				componentAccessor.customFieldManager.getCustomFieldObject(startDateCustomFieldFullId).name :
				componentAccessor.i18nHelper.getText("issue.field.created")
	}

	public String getEndDateLabel() {
		return (endDateCustomFieldFullId != null) ?
				componentAccessor.customFieldManager.getCustomFieldObject(endDateCustomFieldFullId).name :
				componentAccessor.i18nHelper.getText("issue.field.updated")
	}

	/**
	 * Returns the length of the time period spanned by the issues in unit periods, or -1 if undefined.
	 */
	public int getLength() {
		if((earliestDate == null) || (latestDate == null)) {
			return -1
		}

		long length = 1 + ((latestDate.time - earliestDate.time) / millisPerUnitPeriod)
		return length
	}

	public boolean hasSubTasks() {
		return hasSubTasks
	}

	public List getSubTasksByIssue(issue) {
		return subTasksByParentId.get(issue.id)
	}

	public Date getIssueStartDate(issue) {
		return (startDateCustomFieldFullId != null) ? issue.getCustomFieldValue(startDateCustomFieldFullId) : issue.created
	}

	public Date getIssueEndDate(issue) {
		return (endDateCustomFieldFullId != null) ? issue.getCustomFieldValue(endDateCustomFieldFullId) : issue.updated
	}

	/**
	 * Returns the 0-based index of the start date in the range of all spanned dates, or -1 if undefined.
	 */
	public int getIssueStartDateIndex(issue) {
		def startDate = getIssueStartDate(issue)
		long index = (startDate != null) ? ((startDate.time - earliestDate.time) / millisPerUnitPeriod) : -1
		return index
	}

	/**
	 * @see #getIssueStartDateIndex(Task)
	 */
	public int getIssueEndDateIndex(issue) {
		def endDate = getIssueEndDate(issue)
		long index = (endDate != null) ? ((endDate.time - earliestDate.time) / millisPerUnitPeriod) : -1
		return index
	}

	/**
	 * Returns the length of the issue in unit periods, or -1 if undefined.
	 */
	public int getIssueLength(issue) {
		def startIndex = getIssueStartDateIndex(issue)
		def endIndex = getIssueEndDateIndex(issue)
		long length = ((startIndex != -1) && (endIndex != -1)) ? (1 + (endIndex - startIndex)) : -1
		return length
	}

	/**
	 * Returns the number of unit periods beginning from the issue end-date until the latest date, or -1 if undefined.
	 */
	public int getIssueDateFiller(issue) {
		long filler = getLength() - getIssueEndDateIndex(issue) - 1
		return filler
	}

	public int getIssueProgressPercentage(issue) {
		def progressPercentage
		if (progressPercentageCustomFieldFullId != null) {
			progressPercentage = issue.getCustomFieldValue(progressPercentageCustomFieldFullId)
		} else {
			// calculate progress (by comparing "now" to "start date" and "end date")
			def startDate = getIssueStartDate(issue)
			def endDate = getIssueEndDate(issue)
			def difference = ((startDate != null) && (endDate != null)) ? endDate.minus(startDate) : 0
			if (difference > 0) {
				progressPercentage = new Date().minus(startDate) / difference
				progressPercentage = Math.max(Math.min(Math.round(progressPercentage * 100), 100), 0)
			}
		}

		return (progressPercentage != null) ? progressPercentage : 0
	}

	/**
	 * Change this method if you have a different definition of "done".
	 */
	public static boolean isIssueDone(issue) {
		return (issue.resolutionDate != null) // return issue.statusObject.name in [ "Closed", "Done" ]
	}

	public static String getMonthByDate(date) {
		return monthOnlyDateFormat.format(date)
	}

	public static String getDayOfMonthByDate(date) {
		return dayOnlyDateFormat.format(date)
	}

	/**
	 * Returns the midnight moment of the given date.
	 */
	private static trimToMidnight(date) {
		return trimToDay(date, null)
	}

	/**
	 * Returns the midnight moment in the given day of the given date.
	 */
	private static trimToDay(date, dayOfWeek) {
		calendar.setTime(date)
		if(dayOfWeek != null) {
			calendar.set(Calendar.DAY_OF_WEEK, dayOfWeek)
		}
		calendar.set(Calendar.HOUR_OF_DAY, 0)
		calendar.set(Calendar.MINUTE, 0)
		calendar.set(Calendar.SECOND, 0)
		calendar.set(Calendar.MILLISECOND, 0)
		return calendar.time
	}
}
